/**
 * @file dynlist.h
 * @author Mephisto
 * @brief Definitions of dynlist functions and data elements
 */

/**
 * \mainpage Dynlist - by Mephisto
 *
 * <i>Dynlist</i>; the dynamic array library, python style<br>
 * <hr>
 * please check the examples section<br>
 * also, check the dynlist.h documentation
 */

#ifndef DYNLIST_H
#define DYNLIST_H

#include <stdlib.h>

#ifdef THREADSAFE
#include <pthread.h>
#include <semaphore.h>
#endif

/**
 * \brief Dynlist creation shortcut
 *
 * This macro function uses createlist to create a dynlist object
 */
#define dynlist(x/**<name of the list*/, y/**<type of variable to be managed (eg. <i>int</i>)*/) dynlist (x)=createlist(sizeof(y))

/**
 * \brief Dynlist item struct
 *
 * This struct references to the datablock at its index, and to the next and previous item structs
 */
typedef struct item {
	struct item *prev; /**<pointer to previous item struct*/
	void *refer; /**<pointer to datablock*/
	struct item *next; /**<pointer to next item struct*/
}dynlistitem; 

/**
 * \brief Dynlist struct
 *
 * This struct contains state retentive information about the list
 */
typedef struct {
	/**size of individual items*/
	int itemsize;

	/**last requested item index, which is pointed to by <i>last</i>*/
	int lasti;

	/**index of last item in list*/
	int endi;

	/**pointer to last requested item*/
	dynlistitem *last;

	/**pointer to first item in list*/
	dynlistitem *start;

	/**pointer to last item in list*/
	dynlistitem *end;

	#ifdef THREADSAFE
	sem_t lock;
	#endif

}dynlist;

/**
 * \brief Creates a list
 *
 * This function returns a dynlist object for items of size <i>itemsize</i>
 */
dynlist createlist(int itemsize/**<size of items to be managed*/);

/**
 * \brief Create new item in list
 *
 * This function creates a new item in <i>*list</i>
 */
void *newitem(dynlist *list/**<pointer to list*/);

/**
 * \brief Get item pointer
 *
 * This function returns a pointer to the item at index <i>i</i> in list <i>*list</i>
 */
void *getitem(dynlist *list/**<pointer to list*/,
	      int i /**<index*/);

/**
 * \brief Erase item from list
 *
 * This function erases the item at index <i>i</i> from list <i>*list</i>
 */
void eraseitem(dynlist *list/**pointer to list*/, 
			int i/**index*/);

/**
 * \brief Clear list from memory
 *
 * This function clearsthe list <i>*list</i>, and all its items, from memory
 */
void clearlist(dynlist *list/**<pointer to list*/);

/**
 * \brief Get total size
 *
 * This function returns the total size allocated to accomodate <i>list</i>, and all its items
 */
long getsize(dynlist list/**<list*/);

/**
 * \brief Get number of items
 *
 * This function returns the total number of items in <i>list</i>
 */
long getlen(dynlist list/**<list*/);

/**
 * \brief Compile list to static array
 *
 * This function attemps to allocate sufficient space to contain <i>list</i> as a static array, and "compile" the dynamic list into it
 */
void *compile(dynlist list/**list*/);

/**
 * \brief Swap two items in list
 *
 * This function swaps the items at index <i>i</i> and <i>j</i> in-place
 */
void swap(dynlist *list/**<pointer to list*/, 
			int i/**<first index(i)*/, 
			int j/**<second index(j)*/);

/**
 * \brief Reverse list
 *
 * This function reverses all the items in <i>*list</i>
 */
void reverse(dynlist *list/**<pointer to list*/);


/**
 * \brief Add item to list
 *
 * This function adds <i>*item</i> to <i>*list</i> at index <i>i</i>
 */
void additem(dynlist *list/**<pointer to list*/, 
		void *item/**<pointer to previously created item*/,
		int i/**<index at which to add the item*/);

/**
 * \brief Append item to list
 *
 * This function appends <i>*item</i> to <i>*list</i>
 */
void appenditem(dynlist *list/**<pointer to list*/,
		void *item/**<pointer to previously created item*/);

/**
 * \brief Join 2 lists
 *
 * This function appends <i>*list2</i> to <i>*list1</i>
 */
void catlists(dynlist *list1/**<pointer to original list*/,
		dynlist *list2/**<pointer to list to be appended*/);

/**
 * \brief Split a list
 *
 * This function splits <i>*list</i> at <i>i</i> and returns a new list containing anything from <i>i</i>
 */
dynlist splitlist(dynlist *list/**<pointer to original list*/,
		int i/**<index at which to split the list*/);

/**
 * \brief "Decompile" array into dynamic list
 *
 * This function reads an array of <i>num</i> items of size <i>itemsize</i> at location <i>data</i> and converts it to a dynamic list 
 */
dynlist decompile(void *data/**<the array*/, 
			int itemsize/**<size of individual items*/, 
			int num/**<number of items*/);

/**
 * \brief Sort the list, using a user-defined function for comparison
 *
 * This function sorts the items in 
 */
void sortlist(dynlist *list/**<pointer to the list to be sorted*/,
		int (*func)(void*, void*)/**<pointer to comparing function*/);

/**function for internal use*/
dynlistitem *getdl(dynlist *list, int i);

#endif
